    /*

    Copyright (C) 2000-2002 Stefan Westerfeld
                            stefan@space.twc.de

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
   
    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.

    */

#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <map>
#include <vector>
#include <algorithm>
#include <functional>
using namespace std;

#include "common.h"
#include "debug.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <unistd.h>
#include <dirent.h>
#include <strings.h>

using namespace Arts;

namespace Arts {

// TraderRestriction
struct TraderRestriction {
	TraderRestriction(const string& key, const string& value)
		:key(key), value(value) { }

	string key;
	string value;
};

// TraderHelper (internal, no public interface)
class TraderHelper {
public:
    typedef struct InodeStruct
    {
        InodeStruct() : _dev(0), _ino(0) { }
        InodeStruct (dev_t dev, ino_t ino) : _dev(dev), _ino(ino) { }
        ~InodeStruct() { }
        InodeStruct (const InodeStruct& rhs) { *this = rhs; }
        InodeStruct& operator= (const InodeStruct& rhs)
            {
                if (this != &rhs)
               {
                   _dev = rhs._dev;
                   _ino = rhs._ino;
               }
               return (*this);
            }
        inline bool operator< (const InodeStruct& rhs) const
            {
                return ((_dev < rhs._dev) && (_ino < rhs._ino));
            }
           inline bool operator== (const InodeStruct& rhs) const
            {
                return ((_dev == rhs._dev) && (_ino == rhs._ino));
            }
        
        dev_t _dev;
        ino_t _ino;
    } InodeStruct;

    template<class T>
    struct LessStringFunctor : public std::binary_function<T, T, bool>
    {
        inline bool operator() (const T& x, const T& y) const
            {
                return (::strcmp(x.c_str(), y.c_str()) < 0);
            }
    };
    
private:
	std::vector<class TraderOffer_impl *> allOffers;
	static TraderHelper *_instance;
	static std::map<InodeStruct, bool>* _dirMap;
	static std::map<std::string, std::string, TraderHelper::LessStringFunctor<std::string> >* _dirIfaceMap;
	static std::string _emptyString;

protected:
	TraderHelper();
	virtual ~TraderHelper();
	void addDirectory (std::string directory,
                       std::string iface = TraderHelper::_emptyString,
                       std::map<InodeStruct, bool>*& dirsDone = TraderHelper::_dirMap);
	void addDirInterface (std::string directory, std::string interface);
    

public:
	std::vector<TraderOffer>* doQuery(const std::vector<TraderRestriction>& query);
	void load();
	void unload();
	static TraderHelper *the();
	static void shutdown();
};
    std::map<TraderHelper::InodeStruct, bool>* TraderHelper::_dirMap = new std::map<TraderHelper::InodeStruct, bool>();
    std::map<std::string, std::string, TraderHelper::LessStringFunctor<std::string> >*
    TraderHelper::_dirIfaceMap = new std::map<std::string, std::string, TraderHelper::LessStringFunctor<std::string> >();
    std::string TraderHelper::_emptyString = std::string();

// TraderQuery
class TraderQuery_impl : virtual public TraderQuery_skel {
private:
	std::vector<TraderRestriction> restrictions;

public:
	void supports(const std::string& property, const std::string& value);
	std::vector<TraderOffer> *query();
};

REGISTER_IMPLEMENTATION(TraderQuery_impl);

// TraderOffer
class TraderOffer_impl : virtual public TraderOffer_skel {
private:
	std::string _interfaceName;
	std::map<std::string, std::vector<std::string> > property;

public:
	TraderOffer_impl(const std::string& interfaceName,
			 const std::string& filename);

	bool match(const std::vector<TraderRestriction>& restrictions);

	// IDL
	std::string interfaceName();
	std::vector<std::string>* getProperty(const std::string& name);
};

}

// TraderQuery
//----------------------------------------------------------------------------
void TraderQuery_impl::supports(const std::string& property,
                                const std::string& value)
{
	restrictions.push_back(TraderRestriction(property,value));
}

std::vector<TraderOffer> *TraderQuery_impl::query()
{
	return TraderHelper::the()->doQuery(restrictions);
}

// TraderOffer
//----------------------------------------------------------------------------
TraderOffer_impl::TraderOffer_impl(const std::string& interfaceName,
				   const std::string& filename)
   :_interfaceName(interfaceName)
{
	std::ifstream file(filename.c_str());
	std::string line;

	while(getline(file,line))
	{
		std::string key;
		std::vector<std::string> values;

		if(MCOPUtils::tokenize(line,key,values))
			property[key] = values;
	}

	std::vector<std::string>& iname = property["InterfaceName"];
	if (iname.empty())
	    iname.push_back(interfaceName);
}

bool TraderOffer_impl::match(const std::vector<TraderRestriction>& restrictions)
{
	std::vector<TraderRestriction>::const_iterator i;

	for(i = restrictions.begin(); i != restrictions.end();i++)
	{
		const TraderRestriction& res = *i;
		const std::vector<string>& myvalues = property[res.key];
		bool okay = false;

		std::vector<std::string>::const_iterator offerIt = myvalues.begin();
		while(!okay && offerIt != myvalues.end())
		{
			if(res.value == *offerIt)
				okay = true;
			else
				offerIt++;
		}

		if(!okay) return false;
	}
	return true;
}

string TraderOffer_impl::interfaceName()
{
	return _interfaceName;
}

std::vector<std::string>* TraderOffer_impl::getProperty(const std::string& name)
{
	return new vector<string>(property[name]);
}

// TraderHelper
//----------------------------------------------------------------------------

TraderHelper::TraderHelper()
{
	load();
}

TraderHelper::~TraderHelper()
{
	unload();
}

void TraderHelper::load()
{
	const std::vector<std::string> *path = MCOPUtils::traderPath();

	std::vector<std::string>::const_iterator pi;
	std::vector<std::string>::const_iterator pBegin = path->begin();
	std::vector<std::string>::const_iterator pEnd = path->end();
	for (pi = pBegin; pi != pEnd; pi++)
        {
	    arts_debug("TraderHelper::load(): adding directory %s", (*pi).c_str());
	     addDirectory(*pi);
        }
}

void TraderHelper::unload()
{
	std::vector<TraderOffer_impl*>::iterator i;
	std::vector<TraderOffer_impl*>::iterator iBegin = allOffers.begin();
	std::vector<TraderOffer_impl*>::iterator iEnd = allOffers.end();

	for (i = iBegin; i != iEnd; i++)
	{
		TraderOffer_impl *offer = *i;
		offer->_release();
	}
	allOffers.clear();
}

void TraderHelper::addDirInterface (std::string directory,
                                    std::string interface)
{
    if (directory.empty())
        return;
    TraderHelper::_dirIfaceMap->operator[](directory) = interface;
    DIR *dir = ::opendir (directory.c_str());
    if (0 == dir)
        return;

    std::string dirEntry("");
    std::stringstream dirIface;

    struct dirent *de = 0;
    struct stat st;
    struct dirent* dirEntryPtr = (struct dirent *) malloc (sizeof(struct dirent) + size_t(MAXPATHLEN) + sizeof(char));
    while (0 == ::readdir_r(dir, dirEntryPtr, &de) && de)
    {
        if ((::strcmp(dirEntryPtr->d_name, ".") != 0) && (::strcmp(dirEntryPtr->d_name, "..") != 0))
        {
            dirEntry = std::string(directory + "/" + dirEntryPtr->d_name);
            (void) memset (&st, '\0', sizeof (struct stat));
            stat (dirEntry.c_str(), &st);
            if (S_ISDIR (st.st_mode))
            {
                dirIface << interface.c_str() << "::" << dirEntryPtr->d_name;
                arts_debug("addDirInterface(): adding directory '%s' for interface '%s'", dirEntry.c_str(), dirIface.str().c_str());
                addDirInterface (dirEntry, dirIface.str());
		dirEntry = "";
		dirIface.str("");
            }
        }
    }
    closedir (dir);
    free (dirEntryPtr);
}


void TraderHelper::addDirectory(std::string directory,
                                std::string iface,
                                std::map<TraderHelper::InodeStruct, bool>*& dirsDone)
{
    arts_debug("addDirectory(%s,%s)", directory.c_str(), iface.c_str());
    if (0 == TraderHelper::_dirMap)
        TraderHelper::_dirMap = new std::map<InodeStruct, bool>();
    if (0 == dirsDone)
	dirsDone = TraderHelper::_dirMap;
    if (0 == TraderHelper::_dirIfaceMap)
        TraderHelper::_dirIfaceMap =
            new std::map<std::string, std::string, TraderHelper::LessStringFunctor<std::string> >();
    
    struct stat st;
    (void) bzero (&st, sizeof (struct stat));
    stat (directory.c_str(), &st);
    TraderHelper::_dirIfaceMap->operator[](directory) = iface;
	DIR *dir = ::opendir(directory.c_str());
	if (NULL == dir)
	    return;
	static std::string dirEntry("");
	static std::string dirIface("");
    static std::string cIface("");
	struct dirent *de = 0;
	while ((de = ::readdir(dir)) != ((struct dirent *) NULL))
	{
	    if ((::strcmp(de->d_name, ".") != 0) && (::strcmp(de->d_name, "..") != 0))
        {
            dirEntry = std::string(directory + "/" + de->d_name);
            (void) bzero (&st, sizeof (struct stat));
            stat (dirEntry.c_str(), &st);
            if (S_ISDIR (st.st_mode))
            {
                dirIface = de->d_name;
                addDirInterface (dirEntry, dirIface);
            }
        }
	}
    closedir(dir);
    dir = 0;
    std::map<std::string, std::string, TraderHelper::LessStringFunctor<std::string> >::const_iterator dBItr
        = TraderHelper::_dirIfaceMap->begin();
    std::map<std::string, std::string, TraderHelper::LessStringFunctor<std::string> >::const_iterator dEItr
        = TraderHelper::_dirIfaceMap->end();
    std::map<std::string, std::string, TraderHelper::LessStringFunctor<std::string> >::const_iterator dFItr;
    std::map<std::string, std::string, TraderHelper::LessStringFunctor<std::string> >::const_iterator xFItr;
    
    for (dFItr = dBItr; dFItr != dEItr; dFItr++)
    {
        DIR* dir = ::opendir((*dFItr).first.c_str());
        if (NULL == dir)
            continue;
        while ((de = ::readdir(dir)) != ((struct dirent *) NULL))
        {
            arts_debug("addDirectory(): de->d_name = '%s'", de->d_name);
            if ((::strcmp(de->d_name, ".") != 0) && (::strcmp(de->d_name, "..") != 0))
            {
                std::string currentEntry = std::string((*dFItr).first + "/" + de->d_name);
                (void) bzero (&st, sizeof (struct stat));
                stat (currentEntry.c_str(), &st);
                if (S_ISDIR (st.st_mode))
                {
                    xFItr = TraderHelper::_dirIfaceMap->find (currentEntry);
                    if (xFItr != dEItr)
                        arts_debug("'%s' is a directory for interface '%s'", currentEntry.c_str(), (*xFItr).second.c_str());
                    else
                        arts_debug ("ERROR[directory]: directory '%s' does not seem to have any interface assigned!", currentEntry.c_str());
                }
                if (S_ISREG(st.st_mode))
                {
                    if ((::strlen(de->d_name) > 10) &&
                        (::strncmp(&de->d_name[strlen(de->d_name) - 10], ".mcopclass", 10) == 0))
                    {
                            // cut .mcopclass for currentIface;
                        if (!(*dFItr).second.empty())
                        {
                            cIface = (*dFItr).second.c_str();
                            cIface += "::";
                        }
                        else
                            cIface = std::string("");
                        cIface += std::string(de->d_name);
                        cIface = cIface.substr(0, cIface.size() - 10);
                        arts_debug("allOffers.push_back(%s,%s)", cIface.c_str(), currentEntry.c_str());
                        allOffers.push_back(new TraderOffer_impl(cIface, currentEntry));
                    }
                }
            }
        }
        closedir(dir);
    }
    delete (TraderHelper::_dirIfaceMap);
    TraderHelper::_dirIfaceMap = 0;
}

static inline int getOfferPreference(TraderOffer offer)
{
	int result = 0;

	std::vector<std::string> *p = offer.getProperty("Preference");
	if(!p->empty())
		result = atoi(p->front().c_str());
	delete p;

	return result;
}

static bool traderOfferCmp(TraderOffer offer1, TraderOffer offer2)
{
	int p1 = getOfferPreference(offer1);
	int p2 = getOfferPreference(offer2);
	
	if(p1 > p2)
		return true;
	if(p1 < p2)
		return false;

	/* if the Preference= entry doesn't indicate which offer is better,
	 * sort the offers by name to make the trader results at least
	 * independant from the order they are read from the file system */
	return offer1.interfaceName() < offer2.interfaceName();
}

std::vector<TraderOffer> *TraderHelper::doQuery(const std::vector<TraderRestriction>& query)
{
	std:vector<TraderOffer> *result = new std::vector<TraderOffer>;

	std::vector<TraderOffer_impl *>::iterator i;
	for(i = allOffers.begin(); i != allOffers.end(); i++)
	{
		TraderOffer_impl *offer = *i;

		if(offer->match(query))
		{
			result->push_back(TraderOffer::_from_base(offer->_copy()));
		}
	}

	/* sort trader query result by Preference= line for more than one result */
	if(result->size() > 1)
		sort(result->begin(), result->end(), traderOfferCmp);
	return result;
}

TraderHelper *TraderHelper::_instance = 0;
TraderHelper *TraderHelper::the()
{
	if(!_instance) _instance = new TraderHelper();
	return _instance;
}

void TraderHelper::shutdown()
{
	if(_instance)
	{
		delete _instance;
		_instance = 0;
	}
}

namespace Arts {

// trader shutdown
class TraderShutdown :public StartupClass
{
public:
	void startup() { };
	void shutdown() { TraderHelper::shutdown(); };
};

static TraderShutdown traderShutdown;

}

// Dispatcher function for reloading the trader data:
void Dispatcher::reloadTraderData()
{
	TraderHelper::the()->unload();
	TraderHelper::the()->load();
}
